from bauh.api.abstract.model import SuggestionPriority

ALL = {
    'onlyoffice': SuggestionPriority.MEDIUM,
    'gamehub': SuggestionPriority.MEDIUM,
    'rpcs3': SuggestionPriority.MEDIUM,
    'etcher': SuggestionPriority.MEDIUM,
    'postman': SuggestionPriority.LOW,
    'nosqlbooster': SuggestionPriority.LOW,
    'emcas': SuggestionPriority.LOW,
    'azpainter': SuggestionPriority.LOW
}