from bauh.api.abstract.model import SuggestionPriority

ALL = {
    'whatsdesk': SuggestionPriority.TOP,
    'slack': SuggestionPriority.HIGH,
    'discord': SuggestionPriority.HIGH,
    'yakyak': SuggestionPriority.MEDIUM,
    'instagraph': SuggestionPriority.MEDIUM,
    'gimp': SuggestionPriority.MEDIUM,
    'pycharm-professional': SuggestionPriority.LOW,
    'eclipse': SuggestionPriority.LOW,
    'supertuxkart': SuggestionPriority.LOW
}