from bauh.api.abstract.model import SuggestionPriority

ALL = {'google-chrome': SuggestionPriority.HIGH,
       'lutris-git': SuggestionPriority.LOW,
       'xdman': SuggestionPriority.LOW,
       'optimus-manager-qt': SuggestionPriority.LOW}
